local M = {}

----------------------------------------------------------------------------------------------------

local function directoryFileName(mapName, filename)
    --[[
    Parameters:
        mapName  - the name of the map
        filename - the name of the path filename
    Returns:
        <string> - the full directory name for the file
    ]]--
    return "prt/"..mapName.."/"..filename
end

local function load(mapName, filename)
    --[[
    Loads a path from file
    Parameters:
        mapName  - the name of the map the path exists on
        filename - the filename of the path to load
    Returns:
        <graphpath> - the path (nil if the file could not be opened)
    ]]--
    local directoryName = directoryFileName(mapName, filename)

    local file = readFile(directoryName)
    if not file then
        return
    end

    return jsonDecode(file)
end

local function save(mapName, filename, routeLength, waypoints)
    --[[
    Saves a path
    Parameters:
        mapName     - the name of the map the path exists on
        filename    - the name of the file to save the path as
        routeLength - the length of the route
        waypoints   - the route of waypoints to save (should contain their positions and scale)
    ]]--
    serializeJsonToFile(
            directoryFileName(mapName, filename),
            {
                routeLength = routeLength,
                waypoints = waypoints
            },
            true
    )
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.load = load
M.save = save
return M