angular.module('beamng.apps')
  .directive('prtNavigation', ['bngApi', 'StreamsManager', 'Utils', function (bngApi, StreamsManager, Utils) {
    return {
      template: `
      <div style="height: 100%; width: 100%; position: relative;">
        <!-- map container -->

        <div id="overflow-wrap" style=" width: 100%; height: 100%; overflow: hidden">
          <div id="mapContainer" style="overflow: visible;">
            <svg style="overflow: visible">
              <defs>
                <image style="" id="vehicleMarker" width="40" height="40"  x="-20" y="-20" xlink:href="modules/apps/Navigation/vehicleMarker.svg" />
              </defs>
            </svg>
          </div>
          <div style="position: absolute">
            <svg width="40" height="40" style="position: fixed; top:0; left: 50%; margin-top: -20px; margin-left: -20px; transform: scale(1, 1);">
              <!--<path d="M20.272 3.228l-5.733 10.52-3.878 6.51c-1.22 1.87-1.873 4.056-1.877 6.29 0 6.38 5.17 11.55 11.55 11.55 6.38 0 11.55-5.17 11.55-11.55-.002-2.257-.666-4.466-1.91-6.35l-3.92-6.505z" fill="#282828" fill-rule="evenodd" stroke="#fff" stroke-width="2.88"/>-->
              <circle cx="20" cy="20" r="10" stroke="#FFF" stroke-width="1.5" fill="#282828" />
              <text style="line-height:125%" x="14.265" y="1043.581" font-size="14" font-family="sans-serif" letter-spacing="0" word-spacing="0" fill="#fff" transform="translate(1 -1012.362)"><tspan x="14" y="1037">N</tspan></text>
            </svg>
          </div>
        </div>

        <!-- Collectible Display -->
        <div ng-if="collectableTotal > 0" style="font-size: 1.2em; padding: 1%; color: white; background-color: rgba(0, 0, 0, 0.3); position: absolute; top:15px; left: 15px">
          <md-icon style="margin-bottom: 3px;" md-svg-src="{{ 'modules/apps/Navigation/snowman.svg' }}" />
          {{ collectableCurrent + '/' + collectableTotal }}
        </div>

        <style>
          .bounce {
            animation: bounce 1s cubic-bezier(0.4,0.1,0.2,1) both;
          }
          @keyframes bounce {
            0%, 20%, 50%, 80%, 100% { transform: translateY(0); }
            40% { transform: translateY(-4px); }
            60% { transform: translateY(-2px); }
          }
        </style>

      </div>
      `,
      replace: true,
      restrict: 'EA',
      link: function (scope, element, attrs) {
        'use strict';

        let waypoints = {};

        /**
         * Converts a waypoint type into the its colour on the UI
         * @param wpType - the waypoint type
         * @param alpha - the alpha this colour will have
         * @returns {object} the colour of the waypoint
         */
        function waypointTypeToColour(wpType, alpha) {
          console.assert(wpType !== null, "wpType cannot be null");
          console.assert(alpha !== null || alpha === 0, "alpha cannot be null");

          switch (wpType) {
            case "last":
              return "rgba(0, 0, 255, " + alpha + ")";
            case "next":
              return "rgba(255, 0, 0, " + alpha + ")";
            case "nextNext":
              return "rgba(128, 128, 128, " + alpha + ")";
            default:
              console.warn(wpType, "is not a known waypoint type")
          }
        }

        /**
         * Removes a waypoint from the map
         */
        scope.$on("prt_remove_waypoint", function(event, data) {
          waypoints[data.wpName].remove();
        });

        /**
         * Removes all waypoints on the map
         */
        scope.$on("prt_reset", function(event, data) {
          angular.forEach(waypoints, function(_, key) {
            waypoints[key].remove();
          });
        });

        /**
         * Shows a waypoint on the map
         * @param data - should contain:
         *               x - the x coordinate
         *               y - the y coordinate
         *               type - the type of waypoint (next, nextNext or last)
         */
        scope.$on("prt_show_waypoint", function(event, data) {
          waypoints[data.wpName] = hu('<circle>', svg).attr({
            cx: -data.x,
            cy: data.y,
            r: data.scale*4,
            fill: waypointTypeToColour(data.type, 0.4),
            stroke: waypointTypeToColour(data.type, 0.8),
            'stroke-width': 3,
          });
        });



        // See Navigation/app.js

        let root = element[0];
        let mapcontainer = root.children[0].children[0];
        let svg = mapcontainer.children[0];
        let northPointer = root.children[0].children[1].children[0];

        let mapReady = false;
        let rotateMap = true;
        let viewParams = [];
        let mapZoom = 0;
        let mapScale = 1
        let zoomStates = [0, -1000, -2000]
        let zoomSlot = 0;
        // receive live data from the GE map
        let vehicleShapes = {};
        let lastcontrolID = -1;

        let collectableShapes = {};
        // group to store all collectable svgs
        let collGroup;

        let visibilitySlots = [0.2, 0.6, 0.8, 1]
        let activeVisibilitySlot = 1;

        // ability to interact
        element[0].addEventListener('click', function (e) {
          activeVisibilitySlot++;
          if (activeVisibilitySlot >= visibilitySlots.length) activeVisibilitySlot = 0;
          element.css({
            'background-color': 'rgba(50, 50, 50, ' + visibilitySlots[activeVisibilitySlot] + ')',
          })
        });

        element[0].addEventListener('contextmenu', function(e) {
          zoomSlot++;
          mapZoom = zoomSlot < zoomStates.length   ? zoomStates[zoomSlot] : zoomSlot = 0;
        });

        scope.$on('NavigationMapUpdate', function (event, data) {
          if (!mapReady || !data) return;
          updateNavMap(data);

          // center map on thing that is controlled
          centerMap(data.objects[data.controlID]);
        });

        scope.$on('app:resized', function (event, streams) {
          element.css({
            'width': streams.width-25 + "px",
            'height': streams.height-25 + "px",
          })
          setupMap();
        });

        scope.$on('$destroy', function () {
          bngApi.engineLua('extensions.unload("ui_uiNavi")');
          //StreamsManager.remove(requiredStreams);
        });

        let init = false;
        // receive the one-time map setup
        scope.$on('NavigationMap', function (event, data) {
          if (data && !init) {
            setupMap(data);
            init = true;
          }
        });

        scope.$on('CollectablesInit', (event, data) => {
          if (data) setupCollectables(data);
        });

        scope.$on('CollectablesUpdate', (event, data) => {
          if (data) {
            // remove collectable from svg
            collectableShapes[data.collectableName].remove();
            // play animation
            let collectIcon = root.children[1].children[0];
            collectIcon.classList.add('bounce')
            collectIcon.addEventListener("animationend", function () {  // resetting the animation
              collectIcon.classList.remove('bounce');
            });
            // update collected amount
            scope.collectableCurrent = data.collectableAmount;
          };
        });

        function _createCircle(x, y, r, c, s, sw) {
          hu('<circle>', svg).attr({
            cx: x, cy: y, r: 0.5 * r, fill: c, stroke: s, 'stroke-width': sw
          });
        }

        function _createLine(p1, p2, color) {
          hu('<line>', svg).attr({
            x1: p1.x, y1: p1.y, x2: p2.x, y2: p2.y,
            stroke: color,
            strokeWidth: Math.max(p1.radius, p2.radius),
            strokeLinecap: "round",
          });
        }

        function centerMap(obj) {
          let zoom = Math.min(50 + (obj.speed * 3.6) * 1.5, 150);

          // center on what?
          let focusX = -obj.pos[0] / mapScale;
          let focusY = obj.pos[1] / mapScale;

          let borderWidth = root.children[0].clientWidth;
          let borderHeight = root.children[0].clientHeight;
          let degreeNorth = rotateMap ? (obj.rot - 90) : 90;
          let npx = - Math.cos(degreeNorth * Math.PI / 180) * borderWidth * 0.75;
          let npy = borderHeight * 0.5 - Math.sin(degreeNorth * Math.PI / 180) * borderHeight * 0.75;
          let translateX = (((viewParams[0]) + borderWidth/2 - 10) + focusX+10);
          let translateY = (((viewParams[1]) + borderHeight/1.5) + focusY);

          mapcontainer.style.transformOrigin = (((viewParams[0] * -1)) - focusX) + "px " + ((viewParams[1] * -1) - focusY) + "px"
          mapcontainer.style.transform = "translate3d(" + translateX + "px, " + translateY + "px," + mapZoom + "px)" + "rotateX(" + 0 + (zoom / 10) + "deg)" + "rotateZ(" + (180 + Utils.roundDec(obj.rot, 2)) + "deg)"
          //mapcontainer.style.transform = "translate3d(" + translateX + "px, " + translateY + "px," + mapZoom + "px)" + "rotateX(" + 40 + (zoom / 10) + "deg)" + "rotateZ(" + (180 + Utils.roundDec(obj.rot, 2)) + "deg)"

          northPointer.style.transform = 'translate(' + Math.min(Math.max(npx, -borderWidth / 2 - 2), borderWidth / 2) + 'px,' + Math.min(Math.max(npy, 0), borderHeight) + 'px)';
        }

        // no cheating :D
        function hideCollectables(camera) {
          if (camera) {
            collGroup.attr({opacity: 0});
          }
          else {
            collGroup.attr({opacity: 1});
          }
        }

        function updatePlayerShape(key, data) {
          //console.log('updatePlayerShape', key)
          if (vehicleShapes[key]) vehicleShapes[key].remove();
          let isControlled = (key == data.controlID);
          //console.log(data)
          let obj = data.objects[key];

          if (isControlled) {
            if (obj.type == 'Camera') {
              hideCollectables(true);
              vehicleShapes[key] = hu('<circle>', svg);
              vehicleShapes[key].attr('cx', 0);
              vehicleShapes[key].attr('cy', 0);
              vehicleShapes[key].attr('r', 8);
              vehicleShapes[key].css('fill', '#FD6A00');
            }
            else {
              hideCollectables(false);
              vehicleShapes[key] = hu('<use>', svg);
              vehicleShapes[key].attr({ 'xlink:href': '#vehicleMarker' });
            }
          }
          else {
            vehicleShapes[key] = hu('<circle>', svg);
            vehicleShapes[key].attr('cx', 0);
            vehicleShapes[key].attr('cy', 0);
            vehicleShapes[key].attr('r', 10);
            vehicleShapes[key].css('stroke', '#FFFFFF');
            vehicleShapes[key].css('stroke-width', '3px');
            vehicleShapes[key].css('fill', '#A3D39C');
          }
        }

        function updateNavMap(data) {
          // player changed? update shapes?
          if (lastcontrolID != data.controlID) {
            if (lastcontrolID != -1) updatePlayerShape(lastcontrolID, data); // update shape of old vehicle
            updatePlayerShape(data.controlID, data); // update shape of new vehicle
            lastcontrolID = data.controlID;
          }

          // update shape positions
          for (let key in data.objects) {
            let o = data.objects[key];

            // Hack: only show player, since AIs do NOT get re-added when re-spawned
            if (!o.active)
              continue

            if (vehicleShapes[key]) {
              let px = -o.pos[0] / mapScale;
              let py = o.pos[1] / mapScale;
              let rot = Math.floor(-o.rot);
              let iconScale = 1; //Math.min(3, 1 + lastSpeed * 0.5);
              let show = 1;
              if (o.marker == null) {
                o.marker = 'default';
              }
              if (o.marker == 'hidden') {
                show = 0;
              }

              vehicleShapes[key].attr({"transform": "translate(" + px + "," + py + ") scale(" + iconScale + "," + iconScale + ") rotate(" + rot + ")", "opacity": show});
            }
            else {
              updatePlayerShape(key, data);
            }
          }
          // delete missing vehicles
          for (let key in vehicleShapes) {
            if (!data.objects[key]) {
              vehicleShapes[key].remove();
              delete vehicleShapes[key];
            }
          }
        }

        function setupMap(data) {
          if (data != null) {
            element.css({
              'position': 'relative',
              'margin': '10px',
              'perspective': '1000px',
              'background-color': 'rgba(50, 50, 50, 0.6)',
              'border': '2px solid rgba(180, 180, 180, 0.8)',
            });
            svg.style.transform = "scale(-1, -1)"


            if (data.terrainSize) {
              let terrainSizeX = Math.min(data.terrainSize[0] / Math.min(data.squareSize, 1) / mapScale, 2048);
              let terrainSizeY = Math.min(data.terrainSize[1] / Math.min(data.squareSize, 1) / mapScale, 2048);
              viewParams = [
                (-terrainSizeX / 2),
                (-terrainSizeY / 2),
                terrainSizeX,
                terrainSizeY
              ];
            }
            else {
              viewParams = [
                (-512),
                (-512),
                1024,
                1024
              ];
            }

            mapcontainer.style.width = viewParams[2] + "px"
            mapcontainer.style.height = viewParams[3] + "px";
            svg.setAttribute('viewBox', viewParams.join(' '));

            // Draw the map elements
            let minX = 999, maxX = -999;
            let minY = 999, maxY = -999;

            let nodes = data.nodes;

            // figure out dimensions of the road network
            for (let key in nodes) {
              let el = nodes[key];
              if (-el.pos[0] < minX) minX = -el.pos[0];
              if (-el.pos[0] > maxX) maxX = -el.pos[0];
              if (el.pos[1] < minY) minY = el.pos[1];
              if (el.pos[1] > maxY) maxY = el.pos[1];
            }

            // use background image if existing, otherwise draw a simple grid
            if (data.minimapImage && data.terrainOffset && data.terrainSize) {
              // mapcontainer.style.backgroundSize = "100%"
              // mapcontainer.style.backgroundImage = "url('/" + data.minimapImage + "')"
              let bgImage = hu('<image>', svg).attr({
                'x': data.terrainOffset[0] / mapScale,
                'y': data.terrainOffset[1] / mapScale,
                'width': data.terrainSize[0] / mapScale,
                'height': data.terrainSize[1] / mapScale,
                'transform': "scale(-1,-1)",
                'xlink:href': "/" + data.minimapImage,
              });
            } else {
              // draw grid
              let distX = maxX - minX
              let dx = 50
              for (let x = minX; x <= maxX + 1; x += dx) {
                _createLine({ x: x, y: minY, radius: 1 }, { x: x, y: maxY, radius: 1 }, 'rgba(255,255,255,0.3)');
              }
              let distY = maxY - minY
              let dy = 50
              for (let y = minY; y <= maxY + 1; y += dy) {
                _createLine({ x: minX, y: y, radius: 1 }, { x: maxX, y: y, radius: 1 }, 'rgba(255,255,255,0.3)');
              }
            }

            function getDrivabilityColor(d) {
              if (d <= 0.1) return 'rgba(150, 120, 100, 1)'; //'#827864';
              if (d > 0.1 && d < 0.9) return 'rgba(150, 150, 120, 1)'; //'#853800';
              return 'rgba(220, 220, 220, 1)'; //#fa9e28';
            }

            function drawRoads(drivabilityMin, drivabilityMax) {
              for (let key in nodes) {
                let el = nodes[key];
                // walk the links of the node
                if (el.links !== undefined) { // links
                  let d = '';
                  let first = true;
                  for (let key2 in el.links) {
                    let el2 = nodes[key2];
                    let drivability = el.links[key2].drivability;
                    if (drivability >= drivabilityMin && drivability <= drivabilityMax) {
                      _createLine({
                          x: -el.pos[0] / mapScale,
                          y: el.pos[1] / mapScale,
                          radius: Math.min(Math.max(el.radius, 0), 5) * 3
                        }, {
                          x: -el2.pos[0] / mapScale,
                          y: el2.pos[1] / mapScale,
                          radius: Math.min(Math.max(el2.radius, 0), 5) * 3    // prevents massive blobs due to waypoints having larger radius'
                        }, getDrivabilityColor(drivability)
                      );
                    }
                  }

                  /*
                    if(first) {
                      d += 'M' + -el2.pos[0] + ' ' + -el2.pos[1]
                      first = false;
                    } else {
                      d += ' L' + -el2.pos[0] + ' ' + -el2.pos[1]
                    }
                    d +=' Z'
                  let p = hu('<path>', svg).attr({
                    stroke: getDrivabilityColor(drivability),
                    strokeWidth: 10,
                    strokeLinecap: "round",
                    fill: 'none',
                    d: d,
                  });
                    */

                }
              }
            }

            // draw dirt roads and then normal on top
            drawRoads(0, 0.1)
            drawRoads(0.1, 0.9)
            drawRoads(0.9, 1)

            // draw nodes on top
            /*
            let nodeColor = '#fa9e28'
            for (let key in nodes) {
              let el = nodes[key];
              if (el.links !== undefined) {
                nodeColor = getDrivabilityColor(el.links[Object.keys(el.links)[0]]);
              }
              _createCircle(-el.pos[0], el.pos[1], el.radius * 2, nodeColor); // nodes
            }
            */

            mapReady = true;
          }

        }

        // need to draw collectables after roads have been drawn
        function setupCollectables(data) {
          let mapSize = viewParams[2];
          let perimeterScale = 0;

          scope.collectableTotal = data.collectableAmount;
          scope.collectableCurrent = data.collectableCurrent;

          // Calculating the different radius' for the collectible containers
          // based on map size.
          switch (true) {
            case (mapSize <= 1024): {
              perimeterScale = 130;
              break;
            }
            case (mapSize > 1024): {
              perimeterScale = 125
              break;
            }
            case (mapSize >= 2048): {
              perimeterScale = 100;
              break;
            }
          };

          // remove any existing collectable svgs
          for (let key in collectableShapes) {
            collectableShapes[key].remove();
          }
          collGroup = hu('<g>', svg);
          // draw collectable svgs
          for (let item in data.collectableItems) {
            let offset = (Math.random() * 50);
            let coll = hu('<circle>', collGroup).attr({
              cx: (-data.collectableItems[item][1] - (Math.max(Math.random() * 50), 50)), // collectible will be located somewhere within this circle but never in the exact center.
              cy: (data.collectableItems[item][2] - (Math.max(Math.random() * 50), 50)),
              r: perimeterScale,
              fill: 'rgba(255, 160, 0, 0.2)',
              stroke: 'rgba(255, 160, 0, 0.6)',
              'stroke-width': 3,
            });
            // store all collectable svgs
            collectableShapes[item] = coll;
          }
        };

        bngApi.engineLua('extensions.ui_uiNavi.requestUIDashboardMap()');

        bngApi.engineLua(`extensions.core_collectables.sendUIState()`);
      },
    };
  }]);
