angular.module('beamng.apps')
  .directive('prtStartScreen', ['bngApi', function (bngApi) {
    return {
      templateUrl: "modules/pseudoRandomTour/apps/ScenarioStartScreen/templateBMNG.html",
      scope: true,
      replace: true,
      link: function ($scope) {
        'use strict';

        // An enum of overlay screens
        const OVERLAY_SCREENS = Object.freeze(
          {"achievements": 1, "config": 2, "moreInfo": 3, "error": 4, "scenarioCreation": 5}
          );

        // The current overlay screen the user is on (default: null)
        $scope.overlayScreen = null;

        $scope.scenarioSettings = {
          routeLength: null,
          alwaysSunny: null,
          traffic: null,
          waypointSpacing: {
            selected: null,
            options: []
          }
        };

        $scope.scenarioGen = {
          map: null,
          options: []
        }

        $scope.visible = true;
        $scope.mapName = null;

        $scope.achieveStats = {};

        $scope.previousRoute = null;

        $scope.configurations = {
          type : null,
          prompt: null,
          filename: null,
          error: null
        };

        $scope.generated = {
          visible: null,
          numWaypoints: null,
          routeLength: null
        };

        $scope.errorMessage = {
          errorDesc: null,
          img: null
        };

        /**
         * Initialises the default values
         */
        $scope.init = function() {
          $scope.scenarioSettings = {
            alwaysSunny: false,
            traffic: false,
            routeLength: 5000,
            waypointSpacing: {
              selected: "Default",
              options: [
                "Default",
                "Sparse",
                "Sparser"
              ]
            }
          };

          $scope.previousRoute = false;

          $scope.configurations = {
            type: "",
            prompt: "",
            filename: "",
            error: "",
          };

          $scope.errorMessage = {
            errorDesc: "",
            img: ""
          };

          $scope.setMapName();
          $scope.setAchievementStats();
          $scope.findModdedMaps();
          /* Re-check the visibility - this UI will re-appear if the user enters a menu, so set this
             to invisible if the scenario is being played */
          $scope.checkVisibility();

          $scope.previousRouteExists();
        };

        /**
         * Checks and updates the visibility of the UI
         */
        $scope.checkVisibility = function() {
          bngApi.engineLua("extensions.scenario_pseudoRandomTour.isPlayingScenario()", (isVisible) => {
            $scope.$apply(function() {
              $scope.visible = !isVisible;
            });
          });
        };

        /**
         * Finds modded maps and adds them to the possible scenario generation array
         */
        $scope.findModdedMaps = function() {
          bngApi.engineLua("extensions.scenario_pseudoRandomTour.findNewModdedMaps()", (maps) => {
            // If there are modded maps
            if (maps.length > 0) {
              $scope.$apply(function() {
                $scope.scenarioGen = {
                  mapChosen: maps[0].internalName,
                  mapOptions: maps
                }
              });
            }
          });
        };

        /**
         * Converts an achievement level to its colour
         * @param {number} level - the level to convert
         * @returns {string} the colour of the achievement level
         */
        $scope.achievementLevelToColour = function(level) {
          switch(level) {
            case 0:
              return "red";
            case 1:
              return "#CD7F32";
            case 2:
              return "#C0C0C0";
            case 3:
              return "#FFD700";
            case 4:
              return "#B9F2FF";
          }
        };

        /**
         * Setups the achievements statistics
         */
        $scope.setAchievementStats = function() {
          // The width of the progress bar in the table
          const TABLE_WIDTH = 300;

          bngApi.engineLua("extensions.scenario_pseudoRandomTour.getAchievementStats()", (achieveStats) => {
            for (let key in achieveStats) {
              if (achieveStats.hasOwnProperty(key)) {
                let value = achieveStats[key];

                let name = value.name;
                let targets = value.targets;
                let achieveVal = value.value;
                let description = value.description;
                let lastTarget = targets[3];

                // Find the next target
                let nextTarget = 0;
                let currLevel = 0;

                for (let i = 0; i < targets.length; i++) {
                  if (achieveVal < targets[i]) {
                    nextTarget = targets[i];
                    break;
                  } else {
                    currLevel++;
                  }
                }

                let targetStyles = [];
                // Generate the stylings for the achievement targets
                for (let i = 0; i < targets.length; i++) {
                  targetStyles[i] = {
                    "left": TABLE_WIDTH*(targets[i]/targets[3]) + "px",
                    "background-color": $scope.achievementLevelToColour(i+1)
                  }
                }


                // Add in all the necessary details for angular
                $scope.achieveStats[key] = {
                  name: name,
                  style: {
                    "width": (Math.min(1, achieveVal / lastTarget)*100) + "%",
                    "background-color": $scope.achievementLevelToColour(currLevel)
                  },
                  // Show next target or, if they are all achieved, just the current value
                  text: nextTarget > 0 ? achieveVal + "/" + nextTarget : achieveVal,
                  value: achieveVal,
                  targets: {
                    values: targets,
                    styles: targetStyles
                  },
                  desc: description
                }
              }
            }
          });
        };

        /**
         * Sets the map name, ascertained by Lua
         */
        $scope.setMapName = function() {
          bngApi.engineLua("extensions.scenario_pseudoRandomTour.getMapNameFormatted()", (response) => {
            $scope.$apply(function() {
              $scope.mapName = response;
            });
          });
        };

        /**
         * Checks whether a previous route exists
         */
        $scope.previousRouteExists = function() {
          bngApi.engineLua("extensions.scenario_pseudoRandomTour.previousRouteExists()", (response) => {
            $scope.$apply(function() {
              $scope.previousRoute = response;
            });
          });
        };

        /**
         * Updates and shows the generated details on the UI
         * @param routeLength - the length of the route
         * @param numWaypoints - the number of waypoints
         */
        $scope.showGeneratedDetails = function(routeLength, numWaypoints) {
          $scope.generated = {
            visible: true,
            routeLength: routeLength,
            numWaypoints: numWaypoints
          };
        };

        /**
         * Creates the route from the inputted data, handled by Lua
         */
        $scope.apply = function() {
          console.log($scope.scenarioSettings.routeLength + ", " +
            "'" + $scope.scenarioSettings.waypointSpacing.selected.toLowerCase().replace(" ", "_") + "', " +
            + $scope.scenarioSettings.alwaysSunny + ", "
            + $scope.scenarioSettings.traffic);

          bngApi.engineLua("extensions.scenario_pseudoRandomTour.initialise(" +
            $scope.scenarioSettings.routeLength + ", " +
            "'" + $scope.scenarioSettings.waypointSpacing.selected.toLowerCase().replace(" ", "_") + "', " +
            + $scope.scenarioSettings.alwaysSunny + ", "
            + $scope.scenarioSettings.traffic + ")", (response) => {
              let status = response[0];

              if (status === "OK") {
                // Update last generated
                $scope.$apply(function () {
                  $scope.showGeneratedDetails(response[1], response[2]);
                });
              } else {
                if (status === "WARN") {
                  // Update last generated
                  $scope.$apply(function () {
                    $scope.showGeneratedDetails(response[2], response[3]);
                  });
                }

                $scope.showError(status, response[1]);
              }
          });

          $scope.previousRouteExists();

          bngApi.engineLua("extensions.scenario_pseudoRandomTour.freeze()");
        };

        /**
         * Begins the scenario
         */
        $scope.start = function() {
          bngApi.engineLua("extensions.scenario_pseudoRandomTour.start()");

          $scope.visible = false;
        };

        /**
         * Validates a filename
         * @param {string} filename - the filename to validate
         * @return {boolean} if the filename is valid
         */
        $scope.validFilename = function(filename) {
            // Disallowed characters
          return (/^[^\\/:\*\?"<>\|]+$/).test(filename)
            // Not allowed to start with a dot
            && !(/^\./).test(filename)
            // Disallowed filenames
            && !(/^(nul|prn|con|lpt[0-9]|com[0-9])(\.|$)/i).test(filename);
        };

        /**
         * Confirms the settings for the configuration for loading/saving
         */
        $scope.confirmConfigSettings = function() {
          // Check the filename given is valid
          if (!$scope.validFilename($scope.configurations.filename)) {
            $scope.configurations.error = "Invalid filename";
            return;
          }

          let responseHandler = function(response) {
            let status = response[0];
            let ok = status === "OK";

            $scope.$apply(function() {
              if (ok) {
                $scope.closeCurrentWindow();
              } else {
                // Something went wrong, let the user know
                $scope.configurations.error = response[1];
              }
            });

            return ok;
          };

          if ($scope.configurations.type === "Save") {
            bngApi.engineLua("extensions.scenario_pseudoRandomTour.save(" +
              "'" + $scope.configurations.filename + "', " + $scope.generated.routeLength + ")",
                responseHandler);
          } else {
            bngApi.engineLua("extensions.scenario_pseudoRandomTour.load(" +
              "'" + $scope.configurations.filename + "')",
                (response) => {
                  // If the response was OK, show the loaded details for the route
                  if (responseHandler(response)) {
                    $scope.$apply(function() {
                      $scope.showGeneratedDetails(response[1], response[2]);
                    });
                  }
                });
          }
        };

        /**
         * Generates these scenarios for a choosen map
         */
        $scope.generateMap = function() {
          bngApi.engineLua("extensions.scenario_pseudoRandomTour.generateScenario(" +
            "'" + $scope.scenarioGen.mapChosen + "'" +
          ")");

          let mapOptions = $scope.scenarioGen.mapOptions;

          // Remove the created map
          for (let i = 0; i < mapOptions.length; i++) {
            if (mapOptions[i].internalName === $scope.scenarioGen.mapChosen) {
              mapOptions.splice(i, 1);
              break;
            }
          }
        };

        /**
         * Shows the scenario creation overlay
         */
        $scope.showScenarioCreation = function() {
          $scope.overlayScreen = OVERLAY_SCREENS.scenarioCreation;
        };

        /**
         * Shows the additional info overlay
         */
        $scope.showAdditionalInfo = function() {
          $scope.overlayScreen = OVERLAY_SCREENS.moreInfo;
        };

        /**
         * Closes the current window and returns to the main window
         */
        $scope.closeCurrentWindow = function() {
          $scope.overlayScreen = null;
        };

        /**
         * Shows the achievements overlay
         */
        $scope.showAchievements = function() {
          $scope.overlayScreen = OVERLAY_SCREENS.achievements;
        };

        /**
         * Shows an error or warning layer on the UI
         * @param status           - the status (ERR or WARN)
         * @param errorDescription - the error description for the user
         */
        $scope.showError = function(status, errorDescription) {
          $scope.overlayScreen = OVERLAY_SCREENS.error;
          $scope.errorMessage.errorDesc = errorDescription;
          $scope.errorMessage.img = status === "WARN" ? "warning" : "failure";
        };

        /**
         * Opens the configurations window for entering filename
         */
        $scope.openConfigWindow = function() {
          $scope.overlayScreen = OVERLAY_SCREENS.config;
        };

        /**
         * Opens the configuration window with the saving settings
         */
        $scope.openSaveWindow = function() {
          $scope.configurations.type = "Save";
          $scope.configurations.prompt = "Enter the filename to save to:";

          $scope.openConfigWindow();
        };

        /**
         * Opens the configuration window with the loading settings
         */
        $scope.openLoadWindow = function() {
          $scope.configurations.type = "Load";
          $scope.configurations.prompt = "Enter the filename of the path to load:";

          $scope.openConfigWindow();
        };
      },
      // controllerAs: "control"
    };
  }]);
